#!/usr/bin/perl -w
#
# translate parts of the PostgreSQL weekly newsletter
#
# written by Andreas 'ads' Scherbaum <ads@pgug.de>
#
# 2007-09-27:   version 1.0
#               - initial release
# 2008-09-12:   version 2.0
#               - cleanup
#               - multiple languages
# 2009-10-12:   version 3.0
#               - rewrite
#               - use text blocks instead of lines
#               - use text templates

use strict;
use FileHandle;
use Data::Dumper;
$main::version_string = '3.0';    # current version



# handle debugging flag
$main::debug = 0;        # valid values: 0/1

if (defined($ARGV[0]) and $ARGV[0] eq '-d') {
  shift(@ARGV);
  $main::debug = 1;
}


# first (real) parameter must be the configfile
if (!defined($ARGV[2])) {
  help();
  exit(1);
}


# read config file
$main::config = read_config($ARGV[0]);
if (!defined($main::config)) {
  # error message already displayed
  help();
  exit(1);
}
# read translations for local news
%main::translations_local = read_translations($ARGV[1]);
if (!defined(%main::translations_local)) {
  # error message already displayed
  help();
  exit(1);
}
# read translations for product news
%main::translations_products = read_translations($ARGV[2]);
if (!defined(%main::translations_products)) {
  # error message already displayed
  help();
  exit(1);
}

#foreach (sort(keys(%{$main::config}))) {
#  print "$_ : " . $main::config->{$_} . "\n";
#}

$main::url_date = '';

# read in full mail
my $mail = '';
while (my $line = <STDIN>) {
  $mail .= $line;
}

if (length($mail) < 1000) {
  # no mail on STDIN
  print "no mail on STDIN\n";
  help();
  exit(1);
}

# get header and body
my $header = $mail;
$header =~ s/^(.+?\r?\n)\r?\n.*$/$1/s;
my $body = substr($mail, length($header));

# remove signature from end of body
$body =~ s/^(.+?)[\r\n \t]+\-+\(end of broadcast\)\-+.*$/$1/s;


# split header and body
my @header = split(/\r?\n/, $header);
my @body = split(/\r?\n/, $body);


# remove spaces at end of header and body - makes life easier
for (my $i = 0; $i < $#header; $i++) {
  if ($header[$i] =~ /^(.*?)[ \t]+$/) {
    $header[$i] = $1;
  }
}
for (my $i = 0; $i < $#body; $i++) {
  if ($body[$i] =~ /^(.*?)[ \t]+$/) {
    $body[$i] = $1;
  }
}


# add continuous header lines together
my @new_header = ();
foreach my $line (@header) {
  if ($line =~ /^[ \t]/) {
    # the line starts with a space or a tab
    # this is a continuation of the last line
    $new_header[$#new_header] .= "\r\n" . $line;
  } else {
    # new header line
    push(@new_header, $line);
  }
}
@header = @new_header;


# remove all unneeded headers
@new_header = ();
foreach my $line (@header) {
  # skip lines, they will be removed
  if ($line =~ /^X\-Spam\-/) {
  } elsif ($line =~ /^Resent\-/) {
  } elsif ($line =~ /^X\-Greylist/) {
  } elsif ($line =~ /^X\-Virus\-Scanned/) {
  } elsif ($line =~ /^Received:/) {
  } elsif ($line =~ /^List\-/) {
  } elsif ($line =~ /^X\-Mailing\-List/) {
  } elsif ($line =~ /^User\-Agent/) {
  } else {
    # actually use this header
    push(@new_header, $line);
  }
}
@header = @new_header;


# modify/translate the subject line
@new_header = ();
my $orig_subject = '';
my $mail_subject = '';
my $mail_from = '';
# default subject format, can be modified in the config file
my $subject_format = '== PostgreSQL Weekly News - %Y-%m-%d =='; # ISO: YYYY-MM-DD
if (defined($main::config->{'subject_format'})) {
  # replace default format with format from config
  $subject_format = $main::config->{'subject_format'};
}
foreach my $line (@header) {
  if ($line =~ /^Subject: \[ANNOUNCE\] +\=?\=? *Postgre[sS]Q?L? Weekly News \- ([a-zA-Z]+) +([0-9]+) +([0-9]+) *\=?\=?.*$/){
    # save the original subject
    $orig_subject = substr($line, 20);
    #print "saved subject line: |$orig_subject|\n";
    # emulates the modifiers from http://www.php.net/manual/en/function.date.php
    my $r_F = replaced_month_name($1);
    my $r_n = month_number($1);
    my $r_m = sprintf("%02d", $r_n);
    my $r_j = $2;
    my $r_d = sprintf("%02d", $r_j);
    my $r_Y = $3;
    $main::url_date = sprintf("%04d%02d%02d", $r_Y, $r_n, $r_d);

    # replace all modifiers
    $subject_format =~ s/\%F/$r_F/gs;
    $subject_format =~ s/\%n/$r_n/gs;
    $subject_format =~ s/\%m/$r_m/gs;
    $subject_format =~ s/\%j/$r_j/gs;
    $subject_format =~ s/\%d/$r_d/gs;
    $subject_format =~ s/\%Y/$r_Y/gs;

    $line = 'Subject: ' . $subject_format . '';
  }
  push(@new_header, $line);
  if ($line =~ /^Subject: (.*)$/) {
    $mail_subject = $line;
  }
  if ($line =~ /^From: (.*)$/) {
    $mail_from = $line;
  }
}
@header = @new_header;


# add first lines of body
my @new_body = ();
if (defined($main::config->{'orig_article_at'})) {
  push(@new_body, "");
  push(@new_body, $main::config->{'orig_article_at'} . ':');
  # idea: fetch David's blog/rss feed and get the url for this posting
  # new idea: the new url on the main website always contains the current date
  #           build the url based on the date in the subject line
  push(@new_body, "http://www.postgresql.org/community/weeklynews/pwn" . $main::url_date);
}



# re-join body
$body = join("\n", @body);
$body =~ s/^[\r\n ]+//gs;
@body = ();
push(@body, '');
# split again, this time into text parts
while (length($body) > 0) {
  my $next_pos = $body;
  # find next multiple linebreaks
  $next_pos =~ s/^(.*?\n\n+).*$/$1/s;
  if (length($next_pos) == length($body)) {
    # nothing found
    push(@body, $body);
    $body = '';
  } else {
    # match found
    push(@body, $next_pos);
    $body = substr($body, length($next_pos));
  }
}




my $in_products = 0;
my $in_news = 0;
my $in_local = 0;
foreach my $part (@body) {
  #print "seen: |$part|\n";

  if ($part =~ /^\=\= / and $in_products == 1) {
    $in_products = 0;
  }
  if ($part =~ /^\=\= / and $in_news == 1) {
    $in_news = 0;
  }
  if ($part =~ /^\=\= / and $in_local == 1) {
    $in_local = 0;
  }

  if ($in_products == 1) {
    my $tmp_part = $part;
    $tmp_part =~ s/[\r\n]+$//gs;
    foreach my $regex (keys(%main::translations_products)) {
      if ($tmp_part =~ /^$regex$/) {
        my %matches = ();
        if (defined($1)) {$matches{'1'} = $1;}
        if (defined($2)) {$matches{'2'} = $2;}
        if (defined($3)) {$matches{'3'} = $3;}
        if (defined($4)) {$matches{'4'} = $4;}
        if (defined($5)) {$matches{'5'} = $5;}
        $part = $main::translations_products{$regex};
        if (defined($matches{'1'})) {
          $part =~ s/__MATCH_1__/$matches{'1'}/;
        }
        if (defined($matches{'2'})) {
          $part =~ s/__MATCH_2__/$matches{'2'}/;
        }
        if (defined($matches{'3'})) {
          $part =~ s/__MATCH_3__/$matches{'3'}/;
        }
        if (defined($matches{'4'})) {
          $part =~ s/__MATCH_4__/$matches{'4'}/;
        }
        if (defined($matches{'5'})) {
          $part =~ s/__MATCH_5__/$matches{'5'}/;
        }
      }
    }
  }

  if ($in_local == 1) {
    my $tmp_part = $part;
    $tmp_part =~ s/[\r\n]+$//gs;
    foreach my $regex (keys(%main::translations_local)) {
      if ($tmp_part =~ /^$regex$/) {
        my %matches = ();
        if (defined($1)) {$matches{'1'} = $1;}
        if (defined($2)) {$matches{'2'} = $2;}
        if (defined($3)) {$matches{'3'} = $3;}
        if (defined($4)) {$matches{'4'} = $4;}
        if (defined($5)) {$matches{'5'} = $5;}
        $part = $main::translations_local{$regex};
        if (defined($matches{'1'})) {
          $part =~ s/__MATCH_1__/$matches{'1'}/;
        }
        if (defined($matches{'2'})) {
          $part =~ s/__MATCH_2__/$matches{'2'}/;
        }
        if (defined($matches{'3'})) {
          $part =~ s/__MATCH_3__/$matches{'3'}/;
        }
        if (defined($matches{'4'})) {
          $part =~ s/__MATCH_4__/$matches{'4'}/;
        }
        if (defined($matches{'5'})) {
          $part =~ s/__MATCH_5__/$matches{'5'}/;
        }
      }
    }
  }

  if ($in_news == 1) {
    if ($part =~ /^Planet Postgres:(.+)$/s) {
      $part = "Planet PostgreSQL:$1";
    }
    if ($part =~ /^General Bits, Archives and occasional new articles(.*)/s) {
      $part = $main::config->{'translation_news_general_bits'} . $1;
    }
    if ($part =~ /^Postgre[sS]Q?L? Weekly News is brought to you this week by David Fetter(.*)$/s) {
      my $this_translation = $main::config->{'translation_news_authors'};
      my $this_extra = $1;
      my $this_extra_and = $main::config->{'translation_news_authors_and'};
      $this_extra =~ s/\band\b/$this_extra_and/gs;
      $part = $this_translation . $this_extra;
      if ($part =~ /^(.+?[^\.])(\n*)$/s) {
        # complete part with a dot, David usually seems to missing this one
        $part = $1 . '.' . $2;
      }
    }
    if ($part =~ /^Submit news and announcements by Sunday at 3:00pm Pacific time\..*?Please send English language ones to david\@fetter.org, German language.*?to pwn\@pgug.de, Italian language to pwn\@itpug.org./s) {
      $part = $main::config->{'translation_news_new_news_to'};
    }
  }

  if ($part =~ /^\=\= Postgre[sS]Q?L? Weekly News \- .+ \=\=/s) {
    push(@new_body, $subject_format . "\n");

  } elsif ($part =~ /^\=\= Postgre[sS]Q?L? Product News \=\=/s) {
    push(@new_body, '== ' . $main::config->{'translation_product_news'} . ' ==');
    $in_products = 1;

  } elsif ($part =~ /^\=\= Postgre[sS]Q?L? Jobs for ([a-zA-Z]+) \=\=/s) {
    my $this_translation = $main::config->{'translation_jobs'};
    my $month_translation = replaced_month_name($1);
    $this_translation =~ s/__MONTHNAME__/$month_translation/g;
    push(@new_body, '== ' . $this_translation . ' ==');

  } elsif ($part =~ /^\=\= Postgre[sS]Q?L? Local \=\=/s) {
    push(@new_body, '== ' . $main::config->{'translation_local'} . ' ==');
    $in_local = 1;

  } elsif ($part =~ /^\=\= Postgre[sS]Q?L? in the News \=\=/s) {
    push(@new_body, '== ' . $main::config->{'translation_in_the_news'} . ' ==');
    $in_news = 1;

  } elsif ($part =~ /^\=\= Applied Patches \=\=/s) {
    push(@new_body, '== ' . $main::config->{'translation_applied_patches'} . ' ==');

  } elsif ($part =~ /^\=\= Rejected Patches \(for now\) \=\=/s) {
    push(@new_body, '== ' . $main::config->{'translation_rejected_patches'} . ' ==');

  } elsif ($part =~ /^\=\= Pending Patches \=\=/s) {
    push(@new_body, '== ' . $main::config->{'translation_pending_patches'} . ' ==');

  } elsif ($part =~ /^\=\= Postgre[sS]Q?L? (\d+\.\d+) Feature of the Week \=\=/s) {
    my $this_translation = $main::config->{'translation_feature_of_the_week'};
    my $this_version = $1;
    $this_translation =~ s/__VERSION__/$this_version/g;
    push(@new_body, '== ' . $this_translation . ' ==');

  } elsif ($part =~ /^\=\= Postgre[sS]Q?L? Tip of the Week \=\=/s) {
    push(@new_body, '== ' . $main::config->{'translation_tip_of_the_week'} . ' ==');

  } else {
    # seems to be plain text
    push(@new_body, $part);
  }
}


# add signature lines to body
if (defined($main::config->{'signature'})) {
  push(@new_body, $main::config->{'signature'});
}


@body = ();
foreach my $part (@new_body) {
  $part =~ s/\n+$//gs;
  push(@body, $part);
}


# output results
if ($main::debug == 0) {
  my @email_receiver = split(/, ?/, $main::config->{'email_receiver'});
  foreach my $email (@email_receiver) {
    my $fh = new FileHandle;
    if (open($fh, '|-', '/usr/sbin/sendmail -t')) {
      #print $fh join("\r\n", @header) . "\r\n\r\n";
      print $fh $mail_from . "\n";
      print $fh "To: " . $email . "\n";
      print $fh $mail_subject . "\n";
      print $fh "User-Agent: PWN translator v" . $main::version_string . "\n";
      print $fh "\n";
      # FIXME: \n or not?
      print $fh join("\n\n", @body) . "\n";
      close($fh);
    }
  }
}


if ($main::debug == 1) {
  #my $time = time();

  #open(FILE, '>', "/tmp/pwn-" . $time . "-header.txt") || exit(0);
  open(FILE, '>', "/tmp/pwn-header.txt") || exit(0);
  #print FILE join("\r\n", @header) . "\r\n";
  print FILE join("\r\n", @new_header) . "\n";
  close(FILE);

  #open(FILE, '>', "/tmp/pwn-" . $time . "-body.txt") || exit(0);
  open(FILE, '>', "/tmp/pwn-body.txt") || exit(0);
  #print FILE $body;
  print FILE join("\n\n", @body) . "\n";
  close(FILE);
}




# help()
#
# print help page
#
# parameter:
#  none
# return:
#  undefined
sub help {
  print "\n";
  print "$0 [-d] <config file> <local news translations> <product news translations>\n";
  print "\n";
  print "  Translate PostgreSQL weekly newsletter from english into\n";
  print "  another language - at least most static parts\n";
  print "\n";
  print "  Parameter:\n";
  print "   -d writes debugging information to /tmp/pwn-header.txt and /tmp/pwn-body.txt\n";
  print "   - path to config file (see sample included in source package)\n";
  print "   - path to file containing translations for local news (sample included)\n";
  print "   - path to file containing translations for product news (sample included)\n";
  print "\n";
  print "  The mail is feeded through STDIN into the script\n";
  print "\n";
  print "Project website: http://pgfoundry.org/projects/pwn-translation/\n";
  print "\n";
}


# read_config()
#
# read in config file
#
# parameter:
#  - path to config file
# return:
#  - hash with config values
sub read_config {
  my $file = shift;

  if (!-e $file or !-r $file) {
    print STDERR "given config file is not readable!\n";
    return undef;
  }

  # read config
  my %config = ();
  my $last_key = undef;

  my $fh = new FileHandle;
  if (!open($fh, "<", $file)) {
    print STDERR "cannot open config file: $!\n";
    return undef;
  }
  while (my $line = <$fh>) {
    if (length($line) < 3 or substr($line, 0, 1) eq '#') {
      # skip comments and empty lines
      next;
    }
    $line =~ s/[\r\n ]+$//gs;
    if ($line =~ /^([^=\t]+?)[ \t]*=[ \t]*(.*)$/) {
      # valid line
      $config{$1} = $2;
      $last_key = $1;
    } elsif ($line =~ /^\t(.+)$/ and defined($last_key)) {
      $config{$last_key} .= "\n" . $1;
    }
  }
  close($fh);

  # replace placeholders
  my $tab = "\t";
  my $lb = "\n";
  foreach my $key (keys(%config)) {
    $config{$key} =~ s/\\t/$tab/gs;
    $config{$key} =~ s/\\n/$lb/gs;
  }

  return \%config;
}


# read_translations()
#
# read in translations from file
#
# parameter:
#  - path to file containing translations
# return:
#  - hash with translation values
sub read_translations {
  my $file = shift;

  if (!-e $file or !-r $file) {
    print STDERR "given translation file is not readable!\n";
    return undef;
  }

  # read file
  my %translations = ();
  my $in_source = 0;
  my $in_translation = 0;
  my $collected_source = '';
  my $collected_translation = '';
  my $line_nr = 0;

  my $fh = new FileHandle;
  if (!open($fh, "<", $file)) {
    print STDERR "cannot open config file: $!\n";
    return undef;
  }
  while (my $line = <$fh>) {
    $line_nr++;
    if (length($line) < 1) {
      # skip empty lines
      next;
    }
    if ($line =~ /^\-\-\-\-\- source \-\-\-\-\-/) {
      if ($in_source == 1) {
        print STDERR "Already inside 'source' block!\n";
        print STDERR "Line nr: $line_nr\n";
        return undef;
      }
      if (length($collected_source) > 0 and length($collected_translation) > 0) {
        $collected_source =~ s/[\r\n]+$//gs;
        $collected_translation =~ s/[\r\n]+$//gs;
        $translations{$collected_source} = $collected_translation;
      }
      $in_source = 1;
      $in_translation = 0;
      $collected_source = '';
      $collected_translation = '';
    } elsif ($line =~ /^\-\-\-\-\- translation \-\-\-\-\-/) {
      if ($in_translation == 1) {
        print STDERR "Already inside 'translation' block!\n";
        print STDERR "Line nr: $line_nr\n";
        return undef;
      }
      $in_translation = 1;
      $in_source = 0;
    } elsif ($in_source == 1) {
      $collected_source .= $line;
    } elsif ($in_translation == 1) {
      $collected_translation .= $line;
    } else {
      print STDERR "Not inside 'source' or 'translation' block!\n";
      print STDERR "Line nr: $line_nr\n";
      return undef;
    }
  }
  close($fh);
  if (length($collected_source) > 0 and length($collected_translation) > 0) {
    $collected_source =~ s/[\r\n]+$//gs;
    $collected_translation =~ s/[\r\n]+$//gs;
    $translations{$collected_source} = $collected_translation;
  }

  # replace placeholders
  my $tab = "\t";
  my $lb = "\n";
  foreach my $key (keys(%translations)) {
    $translations{$key} =~ s/\\t/$tab/gs;
    $translations{$key} =~ s/\\n/$lb/gs;
  }

  return %translations;
}



# replaced_month_name()
#
# returns the translated month name
#
# parameter:
#  - english month name
# return:
#  - translated month name
sub replaced_month_name {
  my $month = shift;
  my $month_lc = lc($month);

  # replace month name
  if (defined($main::config->{'month:' . $month_lc})) {
    return $main::config->{'month:' . $month_lc};
  }

  return 'undefined month: ' . $month;
}


# month_number()
#
# returns the month number
#
# parameter:
#  - english month name
# return:
#  - month number
sub month_number {
  my $month = shift;
  my $month_lc = lc($month);

  if ($month_lc eq 'january') {return 1;}
  if ($month_lc eq 'february') {return 2;}
  if ($month_lc eq 'march') {return 3;}
  if ($month_lc eq 'april') {return 4;}
  if ($month_lc eq 'may') {return 5;}
  if ($month_lc eq 'june') {return 6;}
  if ($month_lc eq 'july') {return 7;}
  if ($month_lc eq 'august') {return 8;}
  if ($month_lc eq 'september') {return 9;}
  if ($month_lc eq 'october') {return 10;}
  if ($month_lc eq 'november') {return 11;}
  if ($month_lc eq 'december') {return 12;}

  return 0;
}
